/*
 * bbftpc/bbftp_store.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
*/

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/stat.h>
#include <unistd.h>

#include <string.h>
#include <sys/socket.h>

#include "common.h"
#include "client.h"
#include "structures.h"
#include "client_proto.h"


extern  int     transferoption  ; 

extern int  castfd ;

extern xbuffer logmessage;
extern int errcode;


int is_rfio( char *filename ) {
  sSTAT statbuf;
  int   retcode;

  retcode = STAT( filename, &statbuf );

  if( retcode < 0 )
    return( -ERRNO );       /* file error */

  return( 0 == statbuf.st_dev ? 1 : 0 );
}

int bbftp_storeutime( char *filename, struct utimbuf *ftime ) {
    int retcode ;

    retcode = is_rfio( filename );
    if( retcode < 0 ) {
      printxb( &logmessage, "Stat %s : %s", filename, strerror( -retcode ) );
      errcode = 72;
      return( -1 );
    }
    if( 1 == retcode )
      return( 0 );
    retcode = utime(filename,ftime) ;
    if( retcode < 0 ) {
      printxb( &logmessage, "Utime %s : %s", filename, strerror( errno ) );
      errcode = 84;
      return( -1 );
    }
    return( 0 );
}

/******************************************************************************
** bbftp_storeclosecastfile :                                                  *
**                                                                             *
**      Close CASTOR file                                                      *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          castfd                          MODIFIED                           * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftp_storeclosecastfile( char *filename ) {
  if( castfd >= 0 ) {
    int retcode;

    retcode = CLOSE( castfd );

    castfd = -1;
    if( retcode < 0 ) {
      retcode = ERRNO;
      if( filename )
        printxb( &logmessage, "Close %s : %s ", filename, strerror( retcode ) );
      return( -retcode );      
    }
  }
    
  return( 0 );  
}

/*******************************************************************************
** bbftp_storechmod :                                                          *
**                                                                             *
**      Routine to chmod a file                                                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftp_storechmod( char *filename, int mode ) {
    int     retcode ;
    
    retcode = CHMOD( filename, mode );
    
    if ( retcode < 0 ) {
        printxb( &logmessage, "Error chmod on file %s : %s", filename, strerror( ERRNO ) );
        errcode = 85;
        return retcode ;
    }
    return 0 ;
}

/*******************************************************************************
** bbftp_storerename :                                                         *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftp_storerename( char *oldpath, char *newpath ) {
    int     retcode;
    
    retcode = RENAME( oldpath, newpath );
    
    if ( retcode < 0 ) {
        printxb( &logmessage, "Error renaming %s to %s : %s", oldpath, newpath, strerror( ERRNO ));
        errcode = 81;
        return( retcode );
    }
    return( 0 );
}
/*******************************************************************************
** bbftp_storeunlink :                                                         *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftp_storeunlink(char *filename) {
  return( UNLINK( filename ) );
}

/*******************************************************************************
** bbftp_storecheckfile :                                                      *
**                                                                             *
**      Routine to check a file                                                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed no retry                                                *
**           0  OK                                                             *
**           1  Failed                                                         *
**                                                                             *
*******************************************************************************/

int bbftp_storecheckfile( char *filename ) {
	
    sSTAT   statbuf;
    
    int     err;

    if( transferoption & TROPT_UNIX )
      return( 0 );
   
    if( STAT( filename, &statbuf ) < 0 ) {
	int ret;

        /*
        ** It may be normal to get an error if the file
        ** does not exist but some error code must lead
        ** to the interruption of the transfer:
        **        EACCES       : Search permission denied
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG : Path argument too long
        **        ENOTDIR      : A component in path is not a directory
        */
        err = ERRNO;

	switch( err ) {
	  case ENOENT:	/* File doesn't exist */
	    ret = 0;
	  break;
	  case ENOMEM:
	  case EBUSY:
	    ret = 1;
	    break;
	  default:
	    ret = -1;
	    break;
	}

	if( ret != 0 ) {
	  printxb( &logmessage, "Stat(%s): %s", filename, strerror( err ) );
	  errcode = 72;
	}
	return( ret );
    }
   
    /*
    ** The file exists so check if it is a directory
    */
    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
      printxb( &logmessage, "File %s is a directory", filename );
      errcode = 73;
      return( -1 );
    }
    
    return( 0 );
}

/*******************************************************************************
** bbftp_storemkdir :                                                          *
**                                                                             *
**      Routine to create a bunch of directory                                 *
**                                                                             *
**      INPUT variable :                                                       *
**          dirname    :  directory to create    NOT MODIFIED                  *
**          recursif   :  1 recursif             NOT MODIFIED                  *
**                        0 non recursif                                       *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      TO FREE before any return : dirpath                                    *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
int bbftp_storemkdir( char *dirname ) {
    char ch;
    int err;
    int len = strlen( dirname );
    register char *crt;
    char * const end = dirname + len;

    ch = 0;
    crt = end;
    while( -1 == MKDIR( dirname, 0777 ) ) {
      err = ERRNO;
      switch( err ) {
	      
	sSTAT statbuf;

	case EEXIST:
	  if( -1 == STAT( dirname, &statbuf ) || 0 == (statbuf.st_mode & S_IFDIR) ) {
	    err = ENOTDIR;
	    goto mal;
	  }
	  break;
	case ENOENT:
	  if( 0 == (transferoption & TROPT_DIR ) )
	    goto mal;
	  *crt = ch;
	  --crt;
	  if( '/' == *crt )
	    --crt;
	  while( *crt != '/' ) --crt;
	  ch = *crt;
	  *crt = 0;
	  continue;
	mal:
	default:
	  printxb( &logmessage, "Stat(%s): %s", dirname, strerror( err ) );
	  errcode = 83;
	  
	  *crt = ch;
	  return( ENOMEM == err || EBUSY == err ? 1 : -1 );
      }
      break;
    }

    while( crt < end ) {
      *crt = ch;
      ++crt;
      if( 0 == *crt )
	break;
      crt = strchr( crt, '/' );
      if( crt ) {
	*crt = 0;
	ch = '/';
      }
      else
	crt = end;
      
      if( -1 == MKDIR( dirname, 0777 ) ) {
	      
	err = ERRNO;
	goto mal;
      }
    }
    
    return( 0 );
}

/*******************************************************************************
** bbftp_storecreatefile :                                                     *
**                                                                             *
**      Routine to create a file                                               *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      TO FREE before any return : filepath                                   *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
 
int bbftp_storecreatefile( char *filename ) {
    int err;
    int ret;

    if( -1 == UNLINK( filename ) ) {
	    
      ret = 0;
      err = ERRNO;
      switch( err ) {
	char *end;

	case ENOMEM:
        case EBUSY:
          ret = 1;
          break;
	case ENOENT:
	  end = strrchr( filename, '/' );
	  if( NULL == end || filename == end )
	    break;
	  if( transferoption & TROPT_DIR ) {
	    *end = 0;
	    ret = bbftp_storemkdir( filename );
	    *end = '/';
	    if( ret != 0 )
	      return( ret );
	    break;
	  }
	  /* fall through */
	default:
	  ret = -1;
	  break;
      }
      if( ret != 0 ) {
	errcode = 76;
	printxb( &logmessage, "Unlink(%s): %s", filename, strerror( err ) );
	return( ret );
      }
    }

    /*
    ** We create the file
    */
    castfd = OPEN( filename, O_CREAT | O_WRONLY | O_TRUNC, 0666 );

    if( castfd < 0 ) {
        /*
        ** Depending on errno we are going to tell the client to 
        ** retry or not
        */
        err = ERRNO;
        printxb( &logmessage, "Open(%s): %s", filename, strerror( err ) );
	errcode = 77;
	return( ENOMEM == err || EBUSY == err ? 1 : -1 );
    }

    if( transferoption & TROPT_TMP ) {
      int ext;
      
      ext = strlen( filename ) - strlen( ".bbftp.tmp" );

      filename[ext] = 0;

      if( bbftp_storecheckfile( filename ) != 0 )
        return( -1 );

      filename[ext] = '.';
    }
    
    if( 1 == is_rfio( filename ) )
      return( 0 );
	
    /*
    ** And close the file
    */
    ret = bbftp_storeclosecastfile( filename );
    if( ret < 0 ) {
	    
        UNLINK( filename );

	errcode = 80;
	return( ENOSPC == -ret ? -1 : 1 );	
    }

    return( 0 );
}
