/*
 * bbftpd/bbftpd_message.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"

extern int incontrolsock;
extern int outcontrolsock;
extern int recvcontrolto;
extern int sendcontrolto;

extern	int	crypton;
extern	RSA	*myprsa;
extern	RSA	*hisrsa;
extern	unsigned char ciphkey[16];
extern	unsigned char ciphiv[8];
static	int	hisklen = 0;
static	int	hisblklen = 0;
static	int	myklen = 0;
static	int	myblklen = 0;

extern	int	transferoption;

#define	T(x)	f_##x,
handler const handlers[] = {
	HANDLER_LIST
};
#undef  T

#define	T(x)	#x,
const char * const HANDLER_NAMES[] = {
	HANDLER_LIST
};

const char * const MESSAGE_NAMES[] = {
	MESSAGE_LIST
};
#undef  T

const char * const DATA_NAMES[] = {
	"DATA_NOCOMPRESS",
	"DATA_COMPRESS"
};


void recrypt() {
  if( hisrsa != NULL ) {
    RSA_free( hisrsa );
    hisrsa = NULL;
  }
  if( myprsa != NULL ) {
    RSA_free( myprsa );
    myprsa = NULL;
  }
  crypton &= ~2;        /* cipher off */
  hisklen = 0;
  myklen = 0;
  hisblklen = 0;
  myblklen = 0;
}

int readmessage( char *buffer, int len, int to ) {
  int total;

  total = 0;
  while( total < len ) {
    int ret;
    fd_set mask;
    struct timeval timer;
	 
    FD_ZERO( &mask );
    FD_SET( incontrolsock, &mask );
    timer.tv_sec  = to;
    timer.tv_usec = 0;

    ret = select( incontrolsock + 1, &mask, 0, 0, (0 == to ? NULL : &timer) );

    if( -1 == ret ) {
      if( EINTR == errno )
	continue;
      ret = errno; 
      syslog( BBFTPD_ERR, "Readmessage(%d,%d): %s", len, total, strerror( errno ) );
      return( -ret );
    }
    if( 0 == ret ) {
      syslog( BBFTPD_ERR, "Readmessage(%d,%d): %s", len, total, strerror( ETIMEDOUT ) );
      return( -ETIMEDOUT );
    }
	
    ret = recv( incontrolsock, buffer + total, len - total, 0 );
    if( ret < 0 && ENOTSOCK == errno )
      ret = read( incontrolsock, buffer + total, len - total );
		
    if( ret <= 0 ) {
      if( 0 == ret )
	errno = ECONNRESET;
      ret = errno;
     
      if( 0 == (transferoption & TROPT_UNIX) || ECONNRESET != errno ) 
	      syslog( BBFTPD_ERR, "Readmessage(%d,%d): %s", len, total, strerror( errno ) );

      return( -ret );
    }
      
    total += ret;
  }

  /*XXX*//*syslog( BBFTPD_DEBUG, "Readmessage(%d,%d)", len, total );*/
    
  return( 0 );
}

int writemessage( char *buffer, int len, int to ) {
  int total;

  total = 0;
  while( total < len ) {
    int ret;

    fd_set mask;
    struct timeval timer;
     
    FD_ZERO( &mask );
    FD_SET( outcontrolsock, &mask );
    timer.tv_sec  = to;
    timer.tv_usec = 0;

    ret = select( outcontrolsock + 1, 0, &mask, 0, (0 == to ? NULL : &timer) );
    if( -1 == ret ) {
      if( EINTR == errno )
	continue;
      ret = errno;
      syslog( BBFTPD_ERR, "Writemessage(%d,%d): %s", len, total, strerror( errno ) );
      return( -ret );
    }
    if( 0 == ret ) {
      syslog( BBFTPD_ERR, "Writemessage(%d,%d): %s", len, total, strerror( ETIMEDOUT ) );
      return( -ETIMEDOUT );
    }

    ret = send( outcontrolsock, buffer + total, len - total, 0 );
    if( ret < 0 && ENOTSOCK == errno )
      ret = write( outcontrolsock, buffer + total, len - total );
        
    if( ret <= 0 ) {
      if( 0 == ret )
	errno = EPIPE;
      ret = errno;

      if( 0 == (transferoption & TROPT_UNIX) || EPIPE != errno ) 
	      syslog( BBFTPD_ERR, "Writemessage(%d,%d): %s", len, total, strerror( errno ) );

      return( -ret );
    }
        
    total += ret;
  }

  /*XXX*//*syslog( BBFTPD_DEBUG, "Writemessage(%d,%d)", len, total );*/
    
  return( 0 );
}

int readpipe( int p, char *buffer, int len ) {
  int total;

  total = 0;
  while( total < len ) {
    int ret;
    fd_set mask;

    FD_ZERO( &mask );
    FD_SET( p, &mask );

    ret = select( p + 1, &mask, 0, 0, NULL );
    if( ret <= 0 ) {
      if( 0 == ret ) {
	if( 0 == total )
	  total = -1;
	break;
      }
      if( EINTR == errno )
	break;
      total = -1;
      break;
    }
	
    ret = read( p, buffer + total, len - total );
    if( ret <= 0 )
      break;

    total += ret;
  }

  /*XXX*//*syslog( BBFTPD_DEBUG, "Readpipe(%d,%d)", len, total );*/
 
  return( total );
}

int writepipe( int p, char *buffer, int len ) {
  int total;

  total = 0;
  while( total < len ) {
    int ret;
    fd_set mask;

    FD_ZERO( &mask );
    FD_SET( p, &mask );
	
    ret = select( p + 1, 0, &mask, 0, NULL );
    if( ret <= 0 ) {
      if( 0 == ret ) {
	if( 0 == total )
          total = -1;
        break;
      }
      if( EINTR == errno )
	continue;
      total = -1;
      break;
    }
    
    ret = write( p, buffer + total, len - total );
    if( ret <= 0 )
      break;

    total += ret;
  }

  /*XXX*//*syslog( BBFTPD_DEBUG, "Writepipe(%d,%d)", len, total );*/
    
  return( total );
}

int writepackage( int code, int msglen, ... ) {
  struct message header;
  int retcode;
  int cmsglen = msglen;
  int enc = crypton && hisrsa;
  int crtlen;
  static EVP_CIPHER_CTX *ctx = NULL;
  char *fn;
  va_list args;

#ifndef WORDS_BIGENDIAN
  header.code = htonl( code );
  header.msglen = htonl( msglen );
#else
  header.code = code;
  header.msglen = msglen;
#endif

  if( 0 == enc )
    {	    
      retcode = writemessage( (char *)(&header), sizeof( header ), sendcontrolto );
      if( retcode < 0 )
        return( retcode );
    }
  else {
    if( 0 == hisklen )
      hisklen = RSA_size( hisrsa );

    if( 0 == hisblklen && (crypton & 2) ) {
      hisblklen = EVP_CIPHER_block_size( EVP_rc4() ); /*EVP_MAX_BLOCK_LENGTH;*/ 

      if( NULL == ctx ) {
	      ctx = EVP_CIPHER_CTX_new();
	      if( NULL == ctx ) {
		      fn = "EVP_CIPHER_CTX_new";
		      goto ops1;
	      }
      }
      EVP_CIPHER_CTX_init( ctx );
    }
    
    crtlen = (hisblklen ? sizeof( header ) + hisblklen - 1 : hisklen); 
  }
		      
  va_start( args, msglen );
  while( msglen > 0 ) {
    char *buffer = va_arg( args, char * );
    int  length = va_arg( args, int );
    
    if( length > 0 ) {
      msglen -= length;
	    
      if( enc ) {
	if( hisblklen )
	  crtlen += length + hisblklen - 1;
	else {
          crtlen += (length / (hisklen - 12)) * hisklen;
          if( length % (hisklen - 12) )
	    crtlen += hisklen;
	}
      }
      else
      {      
        retcode = writemessage( buffer, length, sendcontrolto );
        if( retcode < 0 ) {
          va_end( args );
          return( retcode );
        }
      }

    }
  }
  va_end( args );
  
  if( enc ) {
    static unsigned char *raw = NULL;
    static int rawlen = 0;
    unsigned char *to;
    struct message ehd;
    int tolen;

    if( hisblklen && hisblklen != 1 )
      crtlen += hisblklen;

    if( crtlen > rawlen ) {
      FREE( raw )
      raw = (unsigned char *)(malloc( crtlen ));
      if( NULL == raw ) {
	rawlen = 0;
	return( -errno );
      }
      rawlen = crtlen;
    }
    
#ifndef WORDS_BIGENDIAN
    ehd.code = htonl( MSG_HIDDEN );
    ehd.msglen = htonl( crtlen );
#else
    ehd.code = MSG_HIDDEN;
    ehd.msglen = crtlen;
#endif

    retcode = writemessage( (char *)(&ehd), sizeof( ehd ), sendcontrolto );
    if( retcode < 0 )
      return( retcode );

    if( hisblklen ) {

      /*EVP_EncryptInit_ex( ctx, EVP_rc4(), NULL, ciphkey, ciphiv );*/
      retcode = EVP_EncryptInit( ctx, EVP_rc4(), ciphkey, ciphiv );
      if( 0 == retcode ) {
	      fn = "EVP_EncryptInit";
	      goto ops1;
      }

      retcode = EVP_EncryptUpdate( ctx, raw, &tolen, (unsigned char *)(&header), sizeof( header ) );
      if( 0 == retcode ) {
	      fn = "EVP_EncryptUpdate";

	ops1:
	      syslog( BBFTPD_ERR, "%s: %s", fn, ERR_error_string( ERR_get_error(), NULL ) );
	      return( -ERR_get_error() );
      }
      
      to = raw + tolen;
    }
    else { 
      retcode = RSA_public_encrypt( sizeof( header ), (unsigned char *)(&header), raw, hisrsa, RSA_PKCS1_PADDING );
      if( -1 == retcode ) {
	      fn = "RSA_public_encrypt";
	      goto ops1;
      }

      to = raw + hisklen;
    }

    va_start( args, msglen );
    while( cmsglen > 0 ) {
      char *buffer = va_arg( args, char * );
      int  length = va_arg( args, int );
    
      if( length > 0 ) {
        cmsglen -= length;

	if( hisblklen ) {
	  retcode = EVP_EncryptUpdate( ctx, to, &tolen, buffer, length );
	  if( 0 == retcode ) {
	    va_end( args );
	    fn = "EVP_EncryptUpdate";
	    goto ops1;
	  }

	  to += tolen;
	}
	else {
          do {
	    int flen;

	    flen = (length > hisklen - 12 ? hisklen - 12 : length);
	    length -= flen;

	    retcode = RSA_public_encrypt( flen, (unsigned char *)(buffer), to, hisrsa, RSA_PKCS1_PADDING );
	    if( -1 == retcode ) {
		    va_end( args );
		    fn = "RSA_public_encrypt";
		    goto ops1;
	    }

	    buffer += flen;
	    to += hisklen;
          }while( length );
	}
	
      }
    }
    va_end( args );

    if( hisblklen ) {

      /*if( 0 == EVP_CipherFinal_ex( ctx, to, &tolen ) )*/
      retcode = EVP_EncryptFinal( ctx, to, &tolen );

      /*XXX*/
      /*EVP_CIPHER_CTX_reset();*/

      if( 0 == retcode ) {
	      fn = "EVP_EncryptFinal";
	      goto ops1;
      }
    }
    
    retcode = writemessage( (char *)(raw), crtlen, sendcontrolto );
    if( retcode < 0 )
      return( retcode );
  }
 
  return( 0 );
}

void reply( int code, char *str ) {
  int len;

  len = strlen( str );
  if( writepackage( code, len, str, len ) < 0 )
    syslog(BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[code] );
}

unsigned char *readpackage( int *length, int *code, unsigned char *bucket ) {
  static struct message header;
  static unsigned char * const empty = "";
  static unsigned char *raw = NULL;
  static int rawlen = 0;
  int withraw;
  unsigned char *final;

  if( readmessage( (char *)(&header), sizeof( header ), recvcontrolto ) < 0 ) {

    none:
    *code = MSG_NONE;
    *length = 0;

    hollow:
    if( NULL == bucket )
      return( empty );
    *(bucket + 2 * sizeof( int )) = 0;
    return( bucket + 2 * sizeof( int ) );
  }
  
#ifndef WORDS_BIGENDIAN
  *code = ntohl( header.code );
  *length = ntohl( header.msglen );
#else
  *code = header.code;
  *length = header.msglen;
#endif
			  
  if( 0 == *length )
    goto hollow;
  
  withraw = ((MSG_HIDDEN == *code && crypton && myprsa) || NULL == bucket);
  
  if( withraw && *length > rawlen ) {
    FREE( raw )
    raw = (unsigned char *)(malloc( *length + 1 ));
    if( NULL == raw ) {
      rawlen = 0;
      *code = MSG_NONE;
      *length = 0;
      return( NULL );
    }
    rawlen = *length;
  }
  
  final = ( withraw ? raw : (bucket + 2 * sizeof( int )) );
  
  if( readmessage( final, *length, recvcontrolto ) < 0 )
	  goto none;

  *(final + *length) = 0;
 
  if( MSG_HIDDEN == *code && crypton && myprsa ) {
    static unsigned char *plain = NULL;
    static int plainlen = 0;
    unsigned char *from, *to;
    unsigned char * const rawend = raw + *length;
    int dlen;
    int crtlen;
    static EVP_CIPHER_CTX *ctx = NULL;
    char *fn;
    int retcode;

    if( 0 == myklen )
      myklen = RSA_size( myprsa );

    if( 0 == myblklen && (crypton & 2) ) {
      myblklen = EVP_CIPHER_block_size( EVP_rc4() ); /*EVP_MAX_BLOCK_LENGTH;*/ 

      if( NULL == ctx ) {
	      ctx = EVP_CIPHER_CTX_new();
	      if( NULL == ctx ) {
		      fn = "EVP_CIPHER_CTX_new";
		      goto ops2;
	      }
      }
      EVP_CIPHER_CTX_init( ctx );
    }

    if( myblklen ) {
      if( 1 == myblklen )
	crtlen = *length;
      else
        crtlen = *length + myblklen - 1 + myblklen;
    }
    else
      crtlen = (*length / myklen) * (myklen - 12);
    
    if( NULL == bucket && crtlen > plainlen ) {
      FREE( plain )
      plain = (unsigned char *)(malloc( crtlen + 1 ));
      if( NULL == plain ) {
	plainlen = 0;
	*code = MSG_NONE;
	*length = 0;
	return( NULL );
      }
      plainlen = crtlen;
    }

    final = ( NULL == bucket ? plain : bucket );

    if( myblklen ) {

      /*EVP_DecryptInit_ex( ctx, EVP_rc4(), NULL, ciphkey, ciphiv );*/
      retcode = EVP_DecryptInit( ctx, EVP_rc4(), ciphkey, ciphiv );
      if( 0 == retcode ) {
	      fn = "EVP_DecryptInit";
	      goto ops2;
      }

      retcode = EVP_DecryptUpdate( ctx, final, &dlen, raw, *length );
      if( 0 == retcode ) {
	fn = "EVP_DecryptUpdate";

	ops2:
	syslog( BBFTPD_ERR, "%s: %s", fn, ERR_error_string( ERR_get_error(), NULL ) );
	goto none;
      }

      /*if( 0 == EVP_CipherFinal_ex( ctx, final + dlen, &dlen ) ) {*/
      retcode = EVP_DecryptFinal( ctx, final + dlen, &dlen );

      /*XXX*/
      /*EVP_CIPHER_CTX_reset( ctx );*/

      if( 0 == retcode ) {
	fn = "EVP_DecryptFinal";
	goto ops2;
      }
    }
    else for( from = raw, to = final; from < rawend; from += myklen, to += dlen ) {
      dlen = RSA_private_decrypt( myklen, from, to, myprsa, RSA_PKCS1_PADDING );
      if( -1 == dlen ) {
	fn = "RSA_private_decrypt";
	goto ops2;
      }
    }
   
    memcpy( (char *)(code), final, sizeof( int ) );
    memcpy( (char *)(length), final + sizeof( int ), sizeof( int ) ); 
#ifndef WORDS_BIGENDIAN
    *code = ntohl( *code );
    *length = ntohl( *length );
#endif

    *(final + 2 * sizeof( int ) + *length) = 0;
    
    return( final + 2 * sizeof( int ) );
  }

  return( final );
}

int handlepackage( int handle ) {
  char	*buffer;
  int	len;
  int	code;
  int	ret;
	    
  buffer = readpackage( &len, &code, NULL );

  if( VALID_MESSAGE( code ) ) {
    syslog( BBFTPD_DEBUG, "%s: got %s message", HANDLER_NAMES[handle], MESSAGE_NAMES[code] );
    switch( code ) {
      case MSG_NONE:
	if( buffer )
	  syslog( BBFTPD_ERR, "%s: Broken connection", HANDLER_NAMES[handle] );
	else {
	  char str[64];
	  
	  sprintf( str, "%s: Memory shortage", HANDLER_NAMES[handle] );
          syslog( BBFTPD_ERR, str );
	  reply( MSG_BAD, str );
	}
        return( -1 );
      default:
        ret = handlers[handle]( buffer, len, code );	/* Goodies here !*/
        if( ret != RET_NONE )
	  return( ret );
	syslog( BBFTPD_ERR, "%s: Unexpected message %s", HANDLER_NAMES[handle], MESSAGE_NAMES[code] );
        return( -1 );
    }
  }
  syslog( BBFTPD_ERR, "%s: Unlisted message %d", HANDLER_NAMES[handle], code );
  return( -1 );
}
