/*
 * bbftpd/bbftpd_store.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include <string.h>
#include <stdio.h>
#include <sys/socket.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "status.h"


extern int  transferoption ;
extern int  requestedstreamnumber ;
extern int  maxstreams ;

extern xbuffer logmessage;

int castfd = -1;


int is_rfio( char *filename ) {
  sSTAT statbuf;
  int   retcode;

  retcode = STAT( filename, &statbuf );

  if( retcode < 0 )
    return( -ERRNO );	/* file error */

  return( 0 == statbuf.st_dev ? 1 : 0 );
}

int closecastfd() {
  if( castfd >= 0 ) {
    int retcode;
	  
    retcode = CLOSE( castfd );
    
    castfd = -1;
    if( retcode < 0 )
      return( -ERRNO );	    
  }
  
  return( 0 );
}  


/*******************************************************************************
** bbftpd_storeutime :                                                         *
**                                                                             *
**      Routine to chage access time a file                                    *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftpd_storeutime( char *filename, struct utimbuf *ftime ) {
    int     retcode ;
  
    retcode = is_rfio( filename );
    if( retcode < 0 ) {
      printxb( &logmessage, "Stat %s : %s ", filename, strerror( -retcode ) );
      return( -1 );
    }
    if( 1 == retcode )
      return( 0 );
    
    retcode = utime( filename, ftime );
    if( retcode < 0 ) {
      printxb( &logmessage, "Utime %s : %s ", filename, strerror( errno ) );
      return( -1 );
    }
    
    return( 0 );
}

/*******************************************************************************
** bbftpd_storechmod :                                                         *
**                                                                             *
**      Routine to chmod a file                                                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftpd_storechmod( char *filename, int mode ) {
    int     retcode ;
  
    retcode = CHMOD( filename, mode );
    
    if( retcode < 0 ) {
      printxb( &logmessage, "Chmod %s : %s ", filename, strerror( ERRNO ) );
      return( -1 );
    }
    
    return( 0 );
}

/*******************************************************************************
** bbftpd_storerename :                                                        *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftpd_storerename( char *oldpath, char *newpath ) {
    int     retcode;
  
    retcode = RENAME( oldpath, newpath );
    
    if( retcode < 0 ) {
      printxb( &logmessage, "Rename %s to %s : %s ", oldpath, newpath, strerror( ERRNO ) );
      return( -1 );
    }

    return( 0 );
}

/*******************************************************************************
** bbftpd_storeunlink :                                                        *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftpd_storeunlink( char *filename ) {
  int retcode;

  retcode = UNLINK( filename );
  
  return( retcode );
}

/*******************************************************************************
** bbftpd_storecheckoptions :                                                  *
**                                                                             *
**      Routine to check the options parameters                                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   POSSIBLY MODIFIED                 * 
**          requestedstreamnumber            POSSIBLY MODIFIED                 *
**                                                                             *
**      RETURN:                                                                *
**          -1  Incorrect options                                              *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

void bbftpd_storecheckoptions() {
	
#ifndef WITH_GZIP
  if( 0 == (transferoption & TROPT_UNIX) ) {
    if( transferoption & TROPT_GZIP )
      syslog( BBFTPD_WARNING, "Compression unavailable in non unix style: option ignored" );
    transferoption &= ~TROPT_GZIP;
  }
#endif

  /* adjust requestedstreamnumber */
  if( requestedstreamnumber > maxstreams ) 
    requestedstreamnumber = maxstreams;
}


static int bbftpd_storecheckfile( char *filepath ) {
  
  sSTAT	statbuf;

  int	err;
  
    if( STAT( filepath, &statbuf ) < 0 ) {
	int ret;

        /*
        ** It may be normal to get an error if the file
        ** does not exist but some error code must lead
        ** to the interruption of the transfer:
        **        EACCES        : Search permission denied
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOTDIR        : A component in path is not a directory
        */
        err = ERRNO;

        switch( err ) {
          case ENOENT:  /* File doesn't exist */
            ret = 0;
            break;
          case ENOMEM:
          case EBUSY:
            ret = 1;
            break;
          default:
            ret = -1;
            break;
        }
		                                                                                
	if( ret != 0 )
	  printxb( &logmessage, "Stat(%s): %s", filepath, strerror( err ) );
			        
	return( ret );
    }

    /*
    ** The file exists so check if it is a directory
    */
    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
      printxb( &logmessage, "File %s is a directory", filepath );
      return( -1 );
    }

   return( 0 ); 
}
 

/*******************************************************************************
** bbftpd_storemkdir :                                                         *
**                                                                             *
**      Routine to create a bunch of directory                                 *
**                                                                             *
**      INPUT variable :                                                       *
**          dirname    :  directory to create    NOT MODIFIED                  *
**          recursif   :  1 recursif             NOT MODIFIED                  *
**                        0 non recursif                                       *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      TO FREE before any return : dirpath                                    *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
int bbftpd_storemkdir( char *dirname ) {
    char ch;
    int err;
    int len = strlen( dirname );
    register char *crt;
    char * const end = dirname + len;

    ch = 0;
    crt = end;
    while( -1 == MKDIR( dirname, 0777 ) ) {
      err = ERRNO;
      switch( err ) {
					                                                                                
        sSTAT statbuf;

        case EEXIST:
          if( -1 == STAT( dirname, &statbuf ) || 0 == (statbuf.st_mode & S_IFDIR) ) {
            err = ENOTDIR;
            goto mal;
          }
          break;
        case ENOENT:
          if( 0 == (transferoption & TROPT_DIR) )
            goto mal;
          *crt = ch;
          --crt;
          if( '/' == *crt )
            --crt;
          while( *crt != '/' ) --crt;
	  ch = *crt;
          *crt = 0;
          continue;
        mal:
        default:
          printxb( &logmessage, "Stat(%s): %s", dirname, strerror( err ) );
	  *crt = ch;
          return( ENOMEM == err || EBUSY == err ? 1 : -1 );
      }
      break;
    }
   
    while( crt < end ) {
      *crt = ch;
      ++crt;
      if( 0 == *crt )
        break;
      crt = strchr( crt, '/' );
      if( crt ) {
        *crt = 0;
        ch = '/';
      }
      else
        crt = end;
                                                                                
      if( -1 == MKDIR( dirname, 0777 ) ) {
	      
        err = ERRNO;
        goto mal;
      }
    }
                                                                                  
    return( 0 );
}
/*******************************************************************************
** bbftpd_storecreatefile :                                                    *
**                                                                             *
**      Routine to create a file                                               *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      TO FREE before any return : filepath                                   *
**                                                                             *
**      RETURN:                                                                *
**          -1  Transfer failed calling has to close the connection no retry   *
**           0  OK  (Does not mean the transfer is successfull                 *
**           1  Transfer failed calling has to close the connection            *
**                                                                             *
*******************************************************************************/
 
int bbftpd_storecreatefile( char *filename ) {
    int err;
    int ret;
	                                                                                
    if( -1 == UNLINK( filename ) ) {
      ret = 0;
      err = ERRNO;
      switch( err ) {
        char *end;

        case ENOMEM:
        case EBUSY:
	  ret = 1;
	  break;
	case ENOENT:
	  end = strrchr( filename, '/' );
	  if( NULL == end || filename == end )
	    break;
	  if( transferoption & TROPT_DIR ) {
	    *end = 0;
	    ret = bbftpd_storemkdir( filename );
            *end = '/';
            if( ret != 0 )
              return( ret );
            break;
	  }
	  /* fall through */
	default:
          ret = -1;
          break;
      }
      if( ret != 0 ) {
        printxb( &logmessage, "Unlink(%s): %s", filename, strerror( err ) );
	return( ret );
      }
    }
    
    /*
    ** We create the file
    */
    castfd = OPEN( filename, O_WRONLY | O_CREAT | O_TRUNC, 0666 );

    if( castfd < 0 ) {
        /*
        ** Depending on errno we are going to tell the client to 
        ** retry or not
        */
        err = ERRNO;
        printxb( &logmessage, "Open(%s): %s ", filename, strerror( err ) );
        /*
        ** We tell the client not to retry in the following case (even in waiting
        ** WAITRETRYTIME the problem will not be solved) :
        **        EACCES        : Search permission denied
        **        EDQUOT        : No more quota 
        **        ENOSPC        : No more space
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOTDIR        : A component in path is not a directory
        */
	return( ENOMEM == err || EBUSY == err ? 1 : -1 );
    }

    if( transferoption & TROPT_TMP ) {
      int ext;

      ext = strlen( filename ) - strlen( ".bbftp.tmp" );
      
      filename[ext] = 0;
     
      if( bbftpd_storecheckfile( filename ) != 0 )
	return( -1 );

      filename[ext] = '.';
    }

    
    if( 1 == is_rfio( filename ) )
      return( 0 );
    
    ret = closecastfd();
    if( ret < 0 ) {
	    
	UNLINK( filename );
	
        printxb( &logmessage, "Close(%s): %s ", filename, strerror( -ret ) );
	return( ENOSPC == -ret ? -1 : 1 );
    }
    
    return( 0 );
}
