/*
 * bbftpd/rsaauth.c
 * Copyright (C) Dan Schrager, in small part, but essential !
 * Dan.Schrager@weizmann.ac.il
 */

/*
 * Author: Tatu Ylonen <ylo@cs.hut.fi>
 * Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
 * Copyright (c) 2000 Markus Friedl.  All rights reserved.
 * Copyright (c) 2000-2002 Todd C. Miller <Todd.Miller@courtesan.com>
 * Copyright (c) 1996 by Internet Software Consortium.
 * Portions Copyright (c) 1995 by International Business Machines, Inc.
 * Copyright (c) 1999 Niels Provos.  All rights reserved.
 */
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <signal.h>
#include <termios.h>
#include <time.h>

#include "structures.h"
#include "daemon_proto.h"
#include "common.h"


char *expand( char *filename ) {
  static char *ef = NULL;
  char *p;
  struct passwd *pass;
  int ul;
  
  if( *filename != '~' )
    return( filename );

  p = strchr( filename, '/' );
  if( NULL == p )
    return( filename );

  ul = p - filename - 1;
  if( ul > 0 ) {
    char *u;
   
    u = (char *)(malloc( ul + 1 ));
    if( NULL == u )
      return( filename );

    memcpy( u, filename + 1, ul );
    *(u + ul) = 0;
    pass = getpwnam( u );
    FREE( u )
  }
  else
    pass = getpwuid( getuid() );

  if( NULL == pass )
    return( filename );

  ef = (char *)(realloc( ef, 1 + strlen( pass->pw_dir ) + strlen( p ) ));
  if( NULL == ef )
    return( filename );

  sprintf( ef, "%s%s", pass->pw_dir, p );
  
  return( ef );
}


void stir() {
  unsigned int seed ;
  struct timespec tp;
  char   buffrand[256];
  int    i;

  OpenSSL_add_all_ciphers();

  clock_gettime( CLOCK_MONOTONIC, &tp );
  seed = tp.tv_nsec;
  srandom(seed);
  for( i=0; i < sizeof( buffrand ); ++i )
    buffrand[i] = random();
  RAND_seed( buffrand, sizeof( buffrand ) );
}


#ifndef HAVE_STRLCPY

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

#endif /* !HAVE_STRLCPY */

#ifndef HAVE_STRLCAT

/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t
strlcat(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}

#endif /* !HAVE_STRLCAT */

#if (!defined(HAVE_B64_NTOP) && !defined(HAVE___B64_NTOP)) || (!defined(HAVE_B64_PTON) && !defined(HAVE___B64_PTON))

/* XXX abort illegal in library */
#define Assert(Cond) if (!(Cond)) abort()

static const char Base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char Pad64 = '=';

/* (From RFC1521 and draft-ietf-dnssec-secext-03.txt)
   The following encoding technique is taken from RFC 1521 by Borenstein
   and Freed.  It is reproduced here in a slightly edited form for
   convenience.

   A 65-character subset of US-ASCII is used, enabling 6 bits to be
   represented per printable character. (The extra 65th character, "=",
   is used to signify a special processing function.)

   The encoding process represents 24-bit groups of input bits as output
   strings of 4 encoded characters. Proceeding from left to right, a
   24-bit input group is formed by concatenating 3 8-bit input groups.
   These 24 bits are then treated as 4 concatenated 6-bit groups, each
   of which is translated into a single digit in the base64 alphabet.

   Each 6-bit group is used as an index into an array of 64 printable
   characters. The character referenced by the index is placed in the
   output string.

                         Table 1: The Base64 Alphabet

      Value Encoding  Value Encoding  Value Encoding  Value Encoding
          0 A            17 R            34 i            51 z
          1 B            18 S            35 j            52 0
          2 C            19 T            36 k            53 1
          3 D            20 U            37 l            54 2
          4 E            21 V            38 m            55 3
          5 F            22 W            39 n            56 4
          6 G            23 X            40 o            57 5
          7 H            24 Y            41 p            58 6
          8 I            25 Z            42 q            59 7
          9 J            26 a            43 r            60 8
         10 K            27 b            44 s            61 9
         11 L            28 c            45 t            62 +
         12 M            29 d            46 u            63 /
         13 N            30 e            47 v
         14 O            31 f            48 w         (pad) =
         15 P            32 g            49 x
         16 Q            33 h            50 y

   Special processing is performed if fewer than 24 bits are available
   at the end of the data being encoded.  A full encoding quantum is
   always completed at the end of a quantity.  When fewer than 24 input
   bits are available in an input group, zero bits are added (on the
   right) to form an integral number of 6-bit groups.  Padding at the
   end of the data is performed using the '=' character.

   Since all base64 input is an integral number of octets, only the
         -------------------------------------------------                       
   following cases can arise:
   
       (1) the final quantum of encoding input is an integral
           multiple of 24 bits; here, the final unit of encoded
	   output will be an integral multiple of 4 characters
	   with no "=" padding,
       (2) the final quantum of encoding input is exactly 8 bits;
           here, the final unit of encoded output will be two
	   characters followed by two "=" padding characters, or
       (3) the final quantum of encoding input is exactly 16 bits;
           here, the final unit of encoded output will be three
	   characters followed by one "=" padding character.
   */

#if !defined(HAVE_B64_NTOP) && !defined(HAVE___B64_NTOP) 
int
b64_ntop(u_char const *src, size_t srclength, char *target, size_t targsize)
{
	size_t datalength = 0;
	u_char input[3];
	u_char output[4];
	int i;

	while (2 < srclength) {
		input[0] = *src++;
		input[1] = *src++;
		input[2] = *src++;
		srclength -= 3;

		output[0] = input[0] >> 2;
		output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
		output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
		output[3] = input[2] & 0x3f;
		Assert(output[0] < 64);
		Assert(output[1] < 64);
		Assert(output[2] < 64);
		Assert(output[3] < 64);

		if (datalength + 4 > targsize)
			return (-1);
		target[datalength++] = Base64[output[0]];
		target[datalength++] = Base64[output[1]];
		target[datalength++] = Base64[output[2]];
		target[datalength++] = Base64[output[3]];
	}
    
	/* Now we worry about padding. */
	if (0 != srclength) {
		/* Get what's left. */
		input[0] = input[1] = input[2] = '\0';
		for (i = 0; i < srclength; i++)
			input[i] = *src++;
	
		output[0] = input[0] >> 2;
		output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
		output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
		Assert(output[0] < 64);
		Assert(output[1] < 64);
		Assert(output[2] < 64);

		if (datalength + 4 > targsize)
			return (-1);
		target[datalength++] = Base64[output[0]];
		target[datalength++] = Base64[output[1]];
		if (srclength == 1)
			target[datalength++] = Pad64;
		else
			target[datalength++] = Base64[output[2]];
		target[datalength++] = Pad64;
	}
	if (datalength >= targsize)
		return (-1);
	target[datalength] = '\0';	/* Returned value doesn't count \0. */
	return (datalength);
}
#endif /* !defined(HAVE_B64_NTOP) && !defined(HAVE___B64_NTOP) */

#if !defined(HAVE_B64_PTON) && !defined(HAVE___B64_PTON)

/* skips all whitespace anywhere.
   converts characters, four at a time, starting at (or after)
   src from base - 64 numbers into three 8 bit bytes in the target area.
   it returns the number of data bytes stored at the target, or -1 on error.
 */

int
b64_pton(char const *src, u_char *target, size_t targsize)
{
	int tarindex, state, ch;
	char *pos;

	state = 0;
	tarindex = 0;

	while ((ch = *src++) != '\0') {
		if (isspace(ch))	/* Skip whitespace anywhere. */
			continue;

		if (ch == Pad64)
			break;

		pos = strchr(Base64, ch);
		if (pos == 0) 		/* A non-base64 character. */
			return (-1);

		switch (state) {
		case 0:
			if (target) {
				if (tarindex >= targsize)
					return (-1);
				target[tarindex] = (pos - Base64) << 2;
			}
			state = 1;
			break;
		case 1:
			if (target) {
				if (tarindex + 1 >= targsize)
					return (-1);
				target[tarindex]   |=  (pos - Base64) >> 4;
				target[tarindex+1]  = ((pos - Base64) & 0x0f)
							<< 4 ;
			}
			tarindex++;
			state = 2;
			break;
		case 2:
			if (target) {
				if (tarindex + 1 >= targsize)
					return (-1);
				target[tarindex]   |=  (pos - Base64) >> 2;
				target[tarindex+1]  = ((pos - Base64) & 0x03)
							<< 6;
			}
			tarindex++;
			state = 3;
			break;
		case 3:
			if (target) {
				if (tarindex >= targsize)
					return (-1);
				target[tarindex] |= (pos - Base64);
			}
			tarindex++;
			state = 0;
			break;
		}
	}

	/*
	 * We are done decoding Base-64 chars.  Let's see if we ended
	 * on a byte boundary, and/or with erroneous trailing characters.
	 */

	if (ch == Pad64) {		/* We got a pad char. */
		ch = *src++;		/* Skip it, get next. */
		switch (state) {
		case 0:		/* Invalid = in first position */
		case 1:		/* Invalid = in second position */
			return (-1);

		case 2:		/* Valid, means one byte of info */
			/* Skip any number of spaces. */
			for (; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					break;
			/* Make sure there is another trailing = sign. */
			if (ch != Pad64)
				return (-1);
			ch = *src++;		/* Skip the = */
			/* Fall through to "single trailing =" case. */
			/* FALLTHROUGH */

		case 3:		/* Valid, means two bytes of info */
			/*
			 * We know this char is an =.  Is there anything but
			 * whitespace after it?
			 */
			for (; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					return (-1);

			/*
			 * Now make sure for cases 2 and 3 that the "extra"
			 * bits that slopped past the last full byte were
			 * zeros.  If we don't check them, they become a
			 * subliminal channel.
			 */
			if (target && target[tarindex] != 0)
				return (-1);
		}
	} else {
		/*
		 * We ended by seeing the end of the string.  Make sure we
		 * have no partial bytes lying around.
		 */
		if (state != 0)
			return (-1);
	}

	return (tarindex);
}

#endif /* !defined(HAVE_B64_PTON) && !defined(HAVE___B64_PTON) */
#endif 


#define GET_32BIT(cp) (((u_long)(u_char)(cp)[0] << 24) | \
                       ((u_long)(u_char)(cp)[1] << 16) | \
                       ((u_long)(u_char)(cp)[2] <<  8) | \
                       ((u_long)(u_char)(cp)[3]))

#define PUT_32BIT(cp, value) do { \
  (cp)[0] = (value) >> 24; \
  (cp)[1] = (value) >> 16; \
  (cp)[2] = (value) >> 8; \
  (cp)[3] = (value); } while (0)
	

u_int
buffer_get_int(u_char **buffer)
{
	u_char buf[4];
	
	memcpy( buf, *buffer, 4 );
	*buffer += 4;
	
	return( GET_32BIT(buf) );
}

/*
 * Returns an arbitrary binary string from the buffer.  The string cannot
 * be longer than 256k.  The returned value points to memory allocated
 * with malloc; it is the responsibility of the calling function to free
 * the data.  If length_ptr is non-NULL, the length of the returned data
 * will be stored there.  A null character will be automatically appended
 * to the returned string, and is not counted in length.
 */
void *
buffer_get_string(u_char **buffer, u_int *length_ptr)
{
	u_char *value;
	u_int len;

	/* Get the length. */
	len = buffer_get_int(buffer);
	if (len > 256 * 1024)
		return( NULL );
	/* Allocate space for the string.  Add one byte for a null character. */
	value = (u_char *)(malloc(len + 1));
	if( NULL == value )
		return( NULL );
	/* Get the string. */
	memcpy( value, *buffer, len );
	*buffer += len;
	/* Append a null character to make processing easier. */
	value[len] = 0;
	/* Optionally return the length of the string. */
	if (length_ptr)
		*length_ptr = len;

	return value;
}

/* XXX does not handle negative BNs */
int
buffer_get_bignum2(u_char **buffer, BIGNUM *value)
{
        u_int len;
        u_char *bin = buffer_get_string(buffer, &len);

        if (NULL == bin || len > 8 * 1024) {
		FREE( bin )
		return( 0 );
	}
        BN_bin2bn(bin, len, value);
        FREE(bin)
	return( 1 );
}

void
buffer_put_int(u_char **buffer, u_int value)
{
        char buf[4];

        PUT_32BIT(buf, value);
	memcpy( *buffer, buf, 4 );
	*buffer += 4;
}

/*
 *  * Stores and arbitrary binary string in the buffer.
 *   */
void
buffer_put_string(u_char **buffer, const void *buf, u_int len)
{
        buffer_put_int(buffer, len);
	memcpy( *buffer, buf, len );
	*buffer += len;
}


#if 0
/*
 * Stores an BIGNUM in the buffer in SSH2 format.
 */
void
buffer_put_bignum2(u_char **buffer, BIGNUM *value)
{
	int bytes = BN_num_bytes(value) + 1;
	u_char *buf = (u_char *)(malloc(bytes));
	int oi;
	int hasnohigh = 0;

	buf[0] = '\0';
	/* Get the value of in binary */
	oi = BN_bn2bin(value, buf+1);
	if (oi != bytes-1)
		fatal("buffer_put_bignum: BN_bn2bin() failed: oi %d != bin_size %d",
		    oi, bytes);
	hasnohigh = (buf[1] & 0x80) ? 0 : 1;
	if (value->neg) {
		/**XXX should be two's-complement */
		int i, carry;
		u_char *uc = buf;
		/*logit("negativ!");*/
		for (i = bytes-1, carry = 1; i>=0; i--) {
			uc[i] ^= 0xff;
			if (carry)
				carry = !++uc[i];
		}
	}
	buffer_put_string(buffer, buf+hasnohigh, bytes-hasnohigh);
	memset(buf, 0, bytes);
	FREE(buf)
}
#endif


int
uuencode(u_char *src, u_int srclength,
    char *target, size_t targsize)
{
	return b64_ntop(src, srclength, target, targsize);
}

int
uudecode(const char *src, u_char *target, size_t targsize)
{
	int len;
	char *encoded, *p;

	/* copy the 'readonly' source */
	encoded = strdup(src);
	/* skip whitespace and data */
	for (p = encoded; *p == ' ' || *p == '\t'; p++)
		;
	for (; *p != '\0' && *p != ' ' && *p != '\t'; p++)
		;
	/* and remove trailing whitespace because __b64_pton needs this */
	*p = '\0';
	len = b64_pton(encoded, target, targsize);
	FREE(encoded)
	return len;
}

void
dump_base64(FILE *fp, u_char *data, u_int len)
{
	char *buf = (char *)(malloc(2*len));
	int i, n;

	if( NULL == buf )
		return;
	
	n = uuencode(data, len, buf, 2*len);
	for (i = 0; i < n; i++) {
		fprintf(fp, "%c", buf[i]);
		if (i % 70 == 69)
			fprintf(fp, "\n");
	}
	if (i % 70 != 69)
		fprintf(fp, "\n");
	FREE(buf)
}


enum types {
	KEY_RSA1,
        KEY_RSA,
        KEY_DSA,
        KEY_UNSPEC
};

RSA *
key_new()
{
	RSA *rsa;
	
	rsa = RSA_new();
	if( NULL != rsa ) {
		BIGNUM *n, *e;

		n = BN_new();
		e = BN_new();
		if( NULL == n || NULL == e ) {
			RSA_free( rsa );
			rsa = NULL;
		}
		else 
			RSA_set0_key( rsa, n, e, NULL );
	}

	return rsa;
}

int
key_equal(RSA *a, RSA *b)
{
	int ret;


	ret = a != NULL && b != NULL;
	if( ret ) {
		const BIGNUM *an, *ae, *bn, *be;

		RSA_get0_key( a, &an, &ae, NULL );
		RSA_get0_key( b, &bn, &be, NULL );

		ret = BN_cmp( an, bn) == 0 && BN_cmp( ae, be) == 0;
	}

	return( ret );
}

int
key_type_from_name(char *name)
{

	if( NULL == name )
	  return( KEY_UNSPEC );
	
        if (strcmp(name, "rsa1") == 0) {
                return KEY_RSA1;
        } else if (strcmp(name, "rsa") == 0) {
                return KEY_RSA;
        } else if (strcmp(name, "dsa") == 0) {
                return KEY_DSA;
        } else if (strcmp(name, "ssh-rsa") == 0) {
                return KEY_RSA;
        } else if (strcmp(name, "ssh-dss") == 0) {
                return KEY_DSA;
        }
        return KEY_UNSPEC;
}

RSA *
key_from_blob(u_char *blob, u_int blen)
{
        u_char *b, *rlen;
        char *ktype;
        int type;
        RSA *key = NULL;

#ifdef DEBUG_PK
        dump_base64(stderr, blob, blen);
#endif
	b = malloc( blen );
	if( NULL == b )
		return( NULL );
        memcpy( b, blob, blen );
        rlen = b;

	ktype = buffer_get_string(&b, NULL);
        type = key_type_from_name(ktype);

	switch (type) {
        case KEY_RSA:
                key = key_new();
		if( NULL == key ) {
			break;
		}
		{
			BIGNUM *n, *e;

			RSA_get0_key( key, (const BIGNUM **)(&n), (const BIGNUM **)(&e), NULL );
			if( 0 == buffer_get_bignum2( &b, e ) || 0 == buffer_get_bignum2( &b, n ) ) {
				RSA_free( key );
				key = NULL;
			}
		}
#ifdef DEBUG_PK
                RSA_print_fp(stderr, key, 8);
#endif
                break;
        default:
                break;
        }

	/*
	**if (key != NULL && b != rlen + blen)
        **        error("key_from_blob: remaining bytes in key blob %d", rlen + blen - b);
	*/
	FREE(ktype)
        FREE(rlen)
        return key;
}



/* returns 1 ok, -1 error */
int
key_read(RSA **ret, char **cpp)
{
	RSA *k;
	char *cp, *space;
	int len, n, type;
	u_char *blob;

	cp = *cpp;

		space = strchr(cp, ' ');
		if (space == NULL) {
			return -1;
		}
		*space = '\0';
		type = key_type_from_name(cp);
		*space = ' ';
		if (type == KEY_UNSPEC) {
			return -1;
		}
		cp = space+1;
		if (*cp == '\0') {
			return -1;
		}
		if (KEY_RSA != type) {
			/* is a key, but different type */
			return -1;
		}
		len = 2*strlen(cp);
		blob = malloc(len);
		if( NULL == blob )
			return( -1 );
		n = uudecode(cp, blob, len);
		if (n < 0) {
			FREE(blob)
			return -1;
		}
		k = key_from_blob(blob, (u_int)n);
		FREE(blob)
		if (k == NULL) {
			return -1;
		}
/*XXX*/
		if (*ret != NULL)
			RSA_free(*ret);
		*ret = k;
#ifdef DEBUG_PK
		RSA_print_fp(stderr, *ret, 8);
#endif
/*XXX*/
		/* advance cp: skip whitespace and data */
		while (*cp == ' ' || *cp == '\t')
			cp++;
		while (*cp != '\0' && *cp != ' ' && *cp != '\t')
			cp++;
		*cpp = cp;
	return 1;
}

u_int
key_size(RSA *k)
{
	const BIGNUM *n;

	RSA_get0_key( k, &n, NULL, NULL );

	return BN_num_bits( n );
}

RSA *
key_from_private(RSA *k)
{
	RSA *n;

	BIGNUM *nn, *ne, *kn, *ke;

	n = key_new();
	if( NULL == n )
		return( NULL );

	RSA_get0_key( n, (const BIGNUM **)(&nn), (const BIGNUM **)(&ne), NULL );
	RSA_get0_key( k, (const BIGNUM **)(&kn), (const BIGNUM **)(&ke), NULL );

	BN_copy( nn, kn);
	BN_copy( ne, ke );

	return n;
}

/* Converts a private to a public key */
RSA *
key_demote(RSA *k)
{
	RSA *pk;

	pk = RSA_new();
	if( NULL != pk ) {
		BIGNUM *kn, *ke;

		RSA_get0_key( k, (const BIGNUM **)(&kn), (const BIGNUM **)(&ke), NULL );

		kn = BN_dup( kn );
		ke = BN_dup( ke );
		if( NULL == kn || NULL == ke ) {
			RSA_free( pk );
			pk = NULL;
		}
		else
			RSA_set0_key( pk, kn, ke, NULL );
	}

	return (pk);
}


int
rsa_public_encrypt(BIGNUM *out, BIGNUM *in, RSA *key)
{
	u_char *inbuf, *outbuf;
	int len, ilen, olen;

	const BIGNUM *n, *e;

	RSA_get0_key( key, &n, &e, NULL );

	if (BN_num_bits( e ) < 2 || !BN_is_odd( e )) {
		return( 0 );
	}

	olen = BN_num_bytes( n );

	outbuf = (u_char *)(malloc(olen));
	if( NULL == outbuf )
		return( -1 );

	ilen = BN_num_bytes(in);
	inbuf = (u_char *)(malloc(ilen));
	if( NULL == inbuf ) {
		FREE( outbuf )
		return( -1 );
	}
	BN_bn2bin(in, inbuf);

	len = RSA_public_encrypt(ilen, inbuf, outbuf, key, RSA_PKCS1_PADDING);
	if (len <= 0) {
		return( -1 );
	}

	BN_bin2bn(outbuf, len, out);

	memset(outbuf, 0, olen);
	memset(inbuf, 0, ilen);
	FREE(outbuf)
	FREE(inbuf)
	return( 1 );
}

int
rsa_private_decrypt(BIGNUM *out, BIGNUM *in, RSA *key)
{
	u_char *inbuf, *outbuf;
	int len, ilen, olen;

	const BIGNUM *n;

	RSA_get0_key( key, &n, NULL, NULL );

	olen = BN_num_bytes( n );

	outbuf = (u_char *)(malloc(olen));
	if( NULL == outbuf )
		return( -1 );

	ilen = BN_num_bytes(in);
	inbuf = (u_char *)(malloc(ilen));
	if( NULL == inbuf ) {
		FREE( outbuf )
		return( -1 );
	}
	BN_bn2bin(in, inbuf);

	len = RSA_private_decrypt(ilen, inbuf, outbuf, key, RSA_PKCS1_PADDING);
	if (len <= 0) {
		return( -1 );
	} else {
		BN_bin2bn(outbuf, len, out);
	}
	memset(outbuf, 0, olen);
	memset(inbuf, 0, ilen);
	FREE(outbuf)
	FREE(inbuf)
	return len;
}


static RSA *
key_load_private_pem(int fd, const char *passphrase,
    char **commentp)
{
	FILE *fp;
	EVP_PKEY *pk = NULL;
	RSA *prv = NULL;
	char *name = "<no key>";

	fp = fdopen(fd, "r");
	if (fp == NULL) {
		/*error("fdopen failed: %s", strerror(errno));*/
		CLOSE(fd);
		return NULL;
	}
	pk = PEM_read_PrivateKey(fp, NULL, NULL, (char *)passphrase);
	if (pk == NULL) {
		(void)ERR_get_error();
	} else {
		prv = EVP_PKEY_get1_RSA(pk);
		if( NULL != prv ) {
			name = "rsa w/o comment";
#ifdef DEBUG_PK
			RSA_print_fp(stderr, prv, 8);
#endif
			if (RSA_blinding_on(prv, NULL) != 1) {
				/*error("key_load_private_pem: RSA_blinding_on failed");*/
				RSA_free(prv);
				prv = NULL;
			}
		} else {
			/*
			**error("PEM_read_PrivateKey: mismatch or "
			**    "unknown EVP_PKEY save_type %d", pk->save_type);
			*/
		}
	}
	FCLOSE(fp);
	if (pk != NULL)
		EVP_PKEY_free(pk);
	if (prv != NULL && commentp)
		*commentp = strdup(name);
	/*
	**debug("read PEM private key done: type %s",
	**    prv ? "RSA" : "<unknown>");
	*/
	return prv;
}

static int
key_perm_ok(int fd, const char *filename)
{
	fSTAT st;

	if (FSTAT(fd, &st) < 0)
		return 0;
	/*
	 * if a key owned by the user is accessed, then we check the
	 * permissions of the file. if the key owned by a different user,
	 * then we don't care.
	 */
#ifdef HAVE_CYGWIN
	if (check_ntsec(filename))
#endif
	if ((st.st_uid == getuid()) && (st.st_mode & 077) != 0) {
		printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
		printf("@         WARNING: UNPROTECTED PRIVATE KEY FILE!          @\n");
		printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
		printf("Permissions 0%3.3o for '%s' are too open.\n",
		    (u_int)st.st_mode & 0777, filename);
		printf("It is recommended that your private key files are NOT accessible by others.\n");
		printf("This private key will be ignored.\n");
		return 0;
	}
	return 1;
}

RSA *
key_load_private(const char *filename, const char *passphrase,
    char **commentp)
{
	RSA *prv;
	int fd;

	fd = OPEN( expand( (char *)(filename) ), O_RDONLY);
	if (fd < 0)
		return NULL;
	if (!key_perm_ok(fd, filename)) {
		CLOSE(fd);
		return NULL;
	}
	/* closes fd */
	prv = key_load_private_pem(fd, passphrase, NULL);
	/* use the filename as a comment for PEM */
	if (commentp)
		*commentp = strdup(filename);
	return prv;
}

int
key_try_load_public(rsal **k, const char *filename, char **commentp)
{
	FILE *f;
	static char line[4096];
	char *cp;
	rsal *b = *k;
	
	f = FOPEN( expand( (char *)(filename) ), "r");
	if (f != NULL) {
		if (commentp)
			*commentp=strdup(filename);
		while (fgets(line, sizeof(line), f)) {
			line[sizeof(line)-1] = '\0';
			cp = line;
			switch (*cp) {
			case '#':
			case '\n':
			case '\0':
				continue;
			}
			/* Skip leading whitespace. */
			for (; *cp && (*cp == ' ' || *cp == '\t'); cp++)
				;
			if (*cp) {
				RSA *r = NULL;
				
				if (key_read(&r, &cp) == 1) {
					rsal *l;
				
					l = (rsal *)(malloc( sizeof( rsal ) ));
					if( NULL == l )
						break;
					l->rsa = r;
					l->next = *k;
					*k = l;
					continue;
				}
			}
		}
		FCLOSE(f);
	}
	if( b != *k ) {
		return( 1 );
	}
	return( 0 );
}

/* load public key from ssh v1 private or any pubkey file */
rsal *
key_load_public(const char *filename, char **commentp)
{
	rsal *pub = NULL;
	char file[MAXPATHLEN];


	/* try ssh2 public key */
	if (key_try_load_public(&pub, filename, commentp) == 1)
		return pub;
	if ((strlcpy(file, filename, sizeof file) < sizeof(file)) &&
	    (strlcat(file, ".pub", sizeof file) < sizeof(file)) &&
	    (key_try_load_public(&pub, file, commentp) == 1))
		return pub;
	return NULL;
}


BIGNUM *
auth_generate_challenge(RSA *key)
{
	BIGNUM *challenge;
	BN_CTX *ctx;

	if ((challenge = BN_new()) == NULL) {
		return( NULL );
	}
	/* Generate a random challenge. */
	BN_rand(challenge, 256, 0, 0);
	if ((ctx = BN_CTX_new()) == NULL) {
		BN_free( challenge );
		return( NULL );
	}
	{
		const BIGNUM *n;

		RSA_get0_key( key, &n, NULL, NULL );

		BN_mod(challenge, challenge, n, ctx);
	}
	BN_CTX_free(ctx);

	return challenge;
}


#define SSH_RSA_MINIMUM_MODULUS_SIZE      768

int
auth_verify_response(RSA *key, BIGNUM *challenge, u_char response[16])
{
	u_char buf[32], mdbuf[16];
	MD5_CTX md;
	int len;

	{
		const BIGNUM *n;

		RSA_get0_key( key, &n, NULL, NULL );
		
		/* don't allow short keys */
		if (BN_num_bits( n ) < SSH_RSA_MINIMUM_MODULUS_SIZE) {
			/*
			 * error("auth_rsa_verify_response: RSA modulus too small: %d < minimum %d bits",
			 * BN_num_bits(key->n), SSH_RSA_MINIMUM_MODULUS_SIZE);
			 */
			return (0);
		}
	}

	/* The response is MD5 of decrypted challenge */
	len = BN_num_bytes(challenge);
	if (len <= 0 || len > 32) {
		/*
		**fatal("auth_rsa_verify_response: bad challenge length %d", len);
		*/
		return( 0 );
	}
	memset(buf, 0, 32);
	BN_bn2bin(challenge, buf + 32 - len);
	MD5_Init(&md);
	MD5_Update(&md, buf, 32);
	/*
	**MD5_Update(&md, session_id, 16);
	*/
	MD5_Final(mdbuf, &md);

	/* Verify that the response is the original challenge. */
	if (memcmp(response, mdbuf, 16) != 0) {
		/* Wrong answer. */
		return (0);
	}
	/* Correct answer. */
	return (1);
}


void freersalist( rsal *r ) {
  while( r != NULL ) {
    rsal *l;

    RSA_free( r->rsa );
    l = r;
    r = r->next;
    FREE( l )
  }
}


#if 0
int main( int argc, char **argv ) {
  rsal *r;
  char *c = NULL;
  RSA *p;
  RSA *d;
  
  r = key_load_public( argc > 1 ? argv[1] : "/home/dans/.ssh/id_rsa.pub", &c );
 
  printf( "\nmain: Got the public key(s):\n" );
  while( r != NULL ) {
	  rsal *l;
	  
	  RSA_print_fp(stderr, r->rsa, 8);
	  printf( "\nmain: RSA_size is %d\n", RSA_size( r->rsa ) );
	  RSA_free( r->rsa );
	  l = r;
	  r = r->next;
	  FREE( l )
  }

  if( c )
	  printf( "\nmain: commentp is ==>%s<==\n", c );

  c = NULL;
  p = key_load_private( argc > 1 ? argv[1] : "/home/dans/.ssh/id_rsa", "", &c );
  
  printf( "\nmain: Got the private key:\n" );
  if( p != NULL ) {
	RSA_print_fp(stderr, p, 8);
	printf( "\nmain: RSA_size is %d\n", RSA_size( p ) );

	d = key_demote( p );
	if( d != NULL ) {
		RSA_print_fp(stderr, d, 8);
		printf( "\nmain: RSA_size is %d\n", RSA_size( d ) );
		RSA_free( d );
	}
	
	RSA_free( p );
  }
  
  if( c )
	  printf( "\nmain: commentp is ==>%s<==\n", c );

  return( 0 );
}
#endif
